"""Top-level package for pymedio."""

import logging

__title__ = "pymedio"
__description__ = "read arbitrary medical image formats in python"
__url__ = "https://github.com/jcreinhold/pymedio"
__author__ = """Jacob Reinhold"""
__email__ = "jcreinhold@gmail.com"
__version__ = "0.2.1"
__license__ = "MIT"
__copyright__ = "Copyright 2021 Jacob Reinhold"

logging.getLogger(__name__).addHandler(logging.NullHandler())
